﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Windows.Forms;

using RootPro.RootProCAD;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.Geometry;

/// <summary>
/// 表計算アドインプログラム
/// 概要
/// 　表のように並んだ複数の数字の文字図形を選び、足し算や引き算などの計算をした結果を新しい文字図形で配置する RootPro CAD のアドインプログラムです。
/// ※使用許諾
/// ・このアドインを実行した結果についての一切の責任はお使いになられる方にあります。
/// ・このアドインは、プログラムソース自体を改変・流用することが可能です。
/// 　また、このプログラムソースを元に独自のアドインやソフトウェアを作成し、無償・有償を問わず第三者に配布することも認めます。
/// 　ただし、改変・流用する場合は、株式会社ルートプロ(RootPro Corporation)作成の
/// 　当プログラムソースを参考にしていることを第三者にわかるように明記してください。
///
/// 　Copyright @ 2021 RootPro Co.,Ltd. (Japan)
/// </summary>


namespace RCAddInTableCalculationCS
{
    [System.AddIn.AddIn("AppAddIn", Version = "1.0", Publisher = "", Description = "")]
    public partial class AppAddIn
    {
         // アドインプログラム名
        string AddInProgramName = "RootPro CAD";

        // 設定値
        AddInSettings addInSettings = null;

        // ラバーバンド図形リスト
        List<Shape> rubberBandShapes = null;
        int resultShapesVerticalPut = 0;// 結果の文字図形の並べる向き 0:未定,1:縦,2:横(ラバーバンド図形の移動処理用)

        string TableCalculationCommandName = "RootPro.AddIn.TableCalculationCommand";     //コマンド名 

        // エラー番号
        const int RC_ERR_SUCCESS = 1; // エラーなし（成功）
        const int RC_ERR_DIFFERENT_SELECTEDSHAPES_COUNT = -10;// エラー：選択した図形Aと図形Bの数が異なります。
        const int RC_ERR_NON_SELECTEDSHAPES = -11;// エラー：選択図形がありません。
        const int RC_ERR_NOT_CURRENTDRAWINGSHAPE = -12;// エラー：カレント部分図以外の図形が選ばれています。
        const int RC_ERR_SELECTED_NONTEXT = -20;// エラー：文字以外が選ばれています。
        const int RC_ERR_CONVERT_TEXT_TONUM = -21;// エラー：文字を数値に変換できませんでした。
        const int RC_ERR_DIFFERENT_TEXT_PUT = -22;// エラー：選択した図形Aと図形Bの並びの向きが異なります。
        const int RC_ERR_ZERO_DIVISION = -23;   // エラー：選択した図形Bに0があります。(割り算実行時)
        const int RC_ERR_TEXT_LOCATE_ANGLE = -24;// エラー：選択した文字図形に配置角度が0以外のものが含まれています。
        const int RC_ERR_TEXT_DIRECTIONVERTICAL = -25;// エラー：選択した文字図形に縦書きが含まれています。
        const int RC_ERR_TEXT_NOT_ALIGNED = -26;// エラー：選択した文字図形は整列していません。
       
        // コマンドパラメータID
        const int calcModeParameterId = 1;          // 計算モードパラメータID
        const int selectShapeAParameterId = 2;      // 図形AパラメータID
        const int selectShapeBParameterId = 3;      // 図形BパラメータID
        const int coefficientTypeParameterId = 4;   // 係数パラメータID
        const int precisionParameterId = 5;         // 精度パラメータID
        const int roundModeParameterId = 6;         // 丸めモードパラメータID
        const int zeroVisibleParameterId = 7;       // 小数点以下のゼロ表記パラメータID
        const int separateTextTypeParameterId = 8;  // 3桁区切り文字タイプパラメータID
        const int locatePointParameterId = 9;       // 配置点パラメータID

        /// <summary>
        /// 計算モード
        /// </summary>
        enum CalcMode
        {
            CalcMode_Addition = 0,      // 足し算
            CalcMode_Subtraction = 1,   // 引き算
            CalcMode_Multiplication = 2,// 掛け算
            CalcMode_Division = 3,      // 割り算
            CalcMode_Total = 4          // 範囲合計
        }

        /// <summary>
        /// 精度モード
        /// </summary>
        enum PrecisionMode
        {
            PrecisionMode_0 = 0,        // 0
            PrecisionMode_0_0 = 1,      // 0.0
            PrecisionMode_0_00 = 2,     // 0.00
            PrecisionMode_0_000 = 3,    // 0.000
            PrecisionMode_0_0000 = 4,   // 0.0000
            PrecisionMode_0_00000 = 5,  // 0.00000
            PrecisionMode_0_000000 = 6, // 0.000000
            PrecisionMode_F = 7         // 入力値の最大有効桁   
        }

        /// <summary>
        /// 丸めモード
        /// </summary>
        enum RoundMode
        {
            Round_45 = 0,
            Round_Up = 1,
            Round_Cut = 2
        }

        /// <summary>
        /// 小数点以下のゼロ表記
        /// </summary>
        enum ZeroVisible
        {
            ZeroVisible_Show = 0, // あり
            ZeroVisible_Hide = 1  // なし
        }

        /// <summary>
        /// 3桁区切り文字タイプ
        /// </summary>
        enum SeparateTextType
        {
            SeparateTextType_Non = 0,
            SeparateTextType_Comma = 1,
            SeparateTextType_Space = 2
        }

        /// <summary>
        /// 係数タイプ
        /// </summary>
        enum CoefficientType
        {
            CoefficientType_Non = 0,    // なし
            CoefficientType_0_001 = 1,  // 0.001
            CoefficientType_100 = 2     // 100
        }

        // 数値を比較するときに使用する精度
        double RcPrecisionConfusion = Precision.Confusion;
        double RcPrecisionRadianAngular = Precision.RadianAngular;

        /// <summary>
        /// アドイン起動時に呼ばれる処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AppAddIn_Startup(object sender, EventArgs e)
        {
            // ファイルから設定値を読み込む
            if(!LoadAddInSettings(out addInSettings))
                addInSettings = new AddInSettings();// 読み込めなかった場合は初期値

            Debug.Assert(addInSettings != null);

             // 多言語対応リソースのDLLファイルをロード
            AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(ResolveAssembly);

            // コマンドの追加
            CommandManager.AddMacroCommand("RCAddInTableCalculationCS", Properties.Resources.CommandName, MacroCommand);

            Command command = CommandManager.AddCommand(TableCalculationCommandName);// AddIn のコマンド名をセット
            command.Name = Properties.Resources.CommandName;// 日本語や英語での使用者に見せるコマンド名のセット

            // コマンドウィンドウの各行にパラメータを追加
            // メモ：コマンドウィンドウでコマンドのパラメーターを用意する
            // 各パラメーターは ParameterItemType がありそれぞれ扱いたい値によって使い分ける
            // パラメータの種類
            // ParameterItemType.List       --- PointParameterItem;      // 座標パラメータ         -- 座標値を入力できます。マウスで図面上から指定することも可能です。
            // ParameterItemType.PointList  --- PointListParameterItem;  // 点列パラメータ         -- 連続した複数の座標値を入力できます。
            // ParameterItemType.Length     --- LengthParameterItem;     // 長さパラメータ         -- 長さを入力できます。パラメータがカレントになると角度を入力するための補助入力パラメータも使えます。
            // ParameterItemType.Angle      --- AngleParameterItem;      // 角度パラメータ         -- 角度を入力できます。パラメータがカレントになると角度を入力するための補助入力パラメータも使えます。
            // ParameterItemType.Double     --- DoubleParameterItem;     // 倍精度浮動小数点数パラメータ
            // ParameterItemType.Integer    --- IntegerParameterItem;    // 整数パラメータ
            // ParameterItemType.Scale      --- ScaleParameterItem;      // 倍率パラメータ
            // ParameterItemType.Text       --- TextParameterItem;       // 文字パラメータ         -- 文字列を入力できます。
            // ParameterItemType.List       --- ListParameterItem;       // リストパラメータ       -- 選択肢をセットして選ぶことができます。
            // ParameterItemType.Shape      --- ShapeParameterItem;      // 図形パラメータ         -- 図形が選択できます。
            // ParameterItemType.ShapeGroup --- ShapeGroupParameterItem; // 図形グループパラメータ -- 複数の図形が選択できます。
            // ParameterItemType.Multiple   --- MultipleParameterItem;   // 複合パラメータ

            // 第1パラメーター 「計算モード」
            CommandParameterGroup commandParameterGroup1 = command.ParameterGroups.Add(ParameterItemType.List, calcModeParameterId, Properties.Resources.CalcModeParameterName);// 1つ目のパラメータを追加するとパラメータグループが作られ、その中のItemの一つかここで追加したパラメータになる
            ListParameterItem calcModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(calcModeParameterId);// 追加したパラメータのItemを取得
            string itemsString = Properties.Resources.CalcModeListItems;
            string[] items = itemsString.Split(',');
            calcModeListParameterItem.Items = items;
            calcModeListParameterItem.SelectedIndex = addInSettings.calcModeListParameterItemIndex;
            calcModeListParameterItem.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
            calcModeListParameterItem.Description = Properties.Resources.CalcModeListParameterDescription;
            // commandParameterGroup1.Items.Add();// 今のパラメータグループに他の種類のパラメータを追加して、コマンドウィンドウ上のパラメーター名の左右の三角マークをクリックして切り替えることもできる。

            // 第2パラメーター 「図形A」
            command.ParameterGroups.Add(ParameterItemType.ShapeGroup, selectShapeAParameterId, Properties.Resources.ShapeAGroupParameterName);
            ShapeGroupParameterItem shape1ParameterItem = (ShapeGroupParameterItem)command.GetParameterItemByID(selectShapeAParameterId);
            shape1ParameterItem.ShapeTypeFilter = ShapeTypeFilters.Text;
            shape1ParameterItem.Description = Properties.Resources.ShapeAGroupParameterDescription;

            // 第3パラメーター 「図形B」 (計算モードが合計以外のとき)
            if(calcModeListParameterItem.SelectedIndex != (int)CalcMode.CalcMode_Total)
            {
                command.ParameterGroups.Add(ParameterItemType.ShapeGroup, selectShapeBParameterId, Properties.Resources.ShapeBGroupParameterName);
                ShapeGroupParameterItem shape2ParameterItem = (ShapeGroupParameterItem)command.GetParameterItemByID(selectShapeBParameterId);
                shape2ParameterItem.ShapeTypeFilter = ShapeTypeFilters.Text;
                shape2ParameterItem.Description = Properties.Resources.ShapeBGroupParameterDescription;
            }
 
            // 第4パラメーター「係数」
            command.ParameterGroups.Add(ParameterItemType.List, coefficientTypeParameterId, Properties.Resources.CoefficientTypeParameterName);
            ListParameterItem coefficientTypeListParameterItem = (ListParameterItem)command.GetParameterItemByID(coefficientTypeParameterId);
            string itemsString2 = Properties.Resources.CoefficientTypeParameterItems;
            string[] items2 = itemsString2.Split(',');
            coefficientTypeListParameterItem.Items = items2;
            coefficientTypeListParameterItem.SelectedIndex = addInSettings.coefficientListParameterItemIndex;
            coefficientTypeListParameterItem.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
            coefficientTypeListParameterItem.Description = Properties.Resources.CoefficientTypeParameterDescription;
            
            // 第5パラメーター 「精度」
            command.ParameterGroups.Add(ParameterItemType.List, precisionParameterId, Properties.Resources.PrecisionParameterName);
            ListParameterItem precisionListParameterItem = (ListParameterItem)command.GetParameterItemByID(precisionParameterId);
            string itemsString3 = Properties.Resources.PrecisionListParameterItems;
            string[] items3 = itemsString3.Split(',');
            precisionListParameterItem.Items = items3;
            precisionListParameterItem.SelectedIndex = addInSettings.precisionListParameterItemIndex;
            precisionListParameterItem.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
            precisionListParameterItem.Description = Properties.Resources.PrecisionListParameterDescription;

            // 第6パラメーター 「丸め」
            command.ParameterGroups.Add(ParameterItemType.List, roundModeParameterId, Properties.Resources.RoundModeParameterName);
            ListParameterItem roundModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(roundModeParameterId);
            string itemsString4 = Properties.Resources.RoundModeListParameterItems;
            string[] items4 = itemsString4.Split(',');
            roundModeListParameterItem.Items = items4;
            roundModeListParameterItem.SelectedIndex = addInSettings.roundModeListParameterItemIndex;
            roundModeListParameterItem.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
            roundModeListParameterItem.Description = Properties.Resources.RoundModeListParameterDescription;

            // 第7パラメーター 「小数点以下のゼロ表記」
            command.ParameterGroups.Add(ParameterItemType.List, zeroVisibleParameterId, Properties.Resources.ZeroVisibleParameterName);
            ListParameterItem zeroVisibleListParameterItem = (ListParameterItem)command.GetParameterItemByID(zeroVisibleParameterId);
            string itemsString5 = Properties.Resources.ZeroVisibleListParameterItems;
            string[] items5 = itemsString5.Split(',');
            zeroVisibleListParameterItem.Items = items5;
            zeroVisibleListParameterItem.SelectedIndex = addInSettings.zeroVisibleListParameterItemIndex;
            zeroVisibleListParameterItem.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
            zeroVisibleListParameterItem.Description = Properties.Resources.ZeroVisibleListParameterDescription;

            // 第8パラメーター 「3桁区切り文字タイプ」
            command.ParameterGroups.Add(ParameterItemType.List, separateTextTypeParameterId, Properties.Resources.SeparateTextTypeParameterName);
            ListParameterItem separateTextTypeListParameterItem = (ListParameterItem)command.GetParameterItemByID(separateTextTypeParameterId);
            string itemsString6 = Properties.Resources.SeparateTextTypeListParameterItems;
            string[] items6 = itemsString6.Split(',');
            separateTextTypeListParameterItem.Items = items6;
            separateTextTypeListParameterItem.SelectedIndex = addInSettings.separateTextTypeListParameterItemIndex;
            separateTextTypeListParameterItem.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
            separateTextTypeListParameterItem.Description = Properties.Resources.SeparateTextTypeListParameterDescription;
            
            // 第9パラメーター「配置点」
            command.ParameterGroups.Add(ParameterItemType.Point, locatePointParameterId, Properties.Resources.LocatePointParameterName);
            PointParameterItem pointParameterItem = (PointParameterItem)command.GetParameterItemByID(locatePointParameterId);
            pointParameterItem.Description = Properties.Resources.LocatePointParameterDescription;

            // 開始イベントと終了イベントの追加
            command.Begin += new CommandBeginEventHandler(command_Begin);
            command.End += new EventHandler(command_End);
        }

        private void AppAddIn_Shutdown(object sender, EventArgs e)
        {
            // 設定値をファイルに書き込む
            Debug.Assert(addInSettings != null);
            SaveAddInSettings(addInSettings);

            CommandManager.RemoveCommand(TableCalculationCommandName);
            CommandManager.RemoveMacroCommand(MacroCommand); // 追加したコマンド関数
        }

        #region VSTA generated code
        private void InternalStartup()
        {
            this.Startup += new System.EventHandler(AppAddIn_Startup);
            this.Shutdown += new System.EventHandler(AppAddIn_Shutdown);
        }
        #endregion


        /// <summary>
        /// 外部フォルダのアセンブリを動的にロード
        /// 多言語対応リソースのDLLファイルをロード
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        private Assembly ResolveAssembly(object sender, ResolveEventArgs args)
        {
            AssemblyName assemblyName = new AssemblyName(args.Name);

            foreach(Assembly assembly in AppDomain.CurrentDomain.GetAssemblies())
            {
                if(AssemblyName.ReferenceMatchesDefinition(assemblyName, assembly.GetName()))
                    return assembly;
            }

            if(assemblyName.Name.EndsWith(".resources"))
            {
                if(assemblyName.CultureName.Equals(Thread.CurrentThread.CurrentUICulture.Name, StringComparison.OrdinalIgnoreCase))
                {
                    string dllPath = Path.GetDirectoryName(this.AddInFullPath);
                    dllPath = Path.Combine(dllPath, assemblyName.CultureName);
                    dllPath = Path.Combine(dllPath, assemblyName.Name + ".dll");

                    if(File.Exists(dllPath))
                    {
                        try
                        {
                            return Assembly.LoadFrom(dllPath);
                        }
                        catch
                        {
                        }
                    }
                }
            }

            return null;
        }


        /// <summary>
        /// マクロコマンドの実行
        /// </summary>
        private void MacroCommand()
        {
            //コマンドの実行
            CommandManager.ExecuteCommand(TableCalculationCommandName);
        }

        /// <summary>
        /// コマンドの開始時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_Begin(Object sender, CommandBeginEventArgs e)
        {
            Command command = (Command)sender;

            // パラメータ変更イベントの追加
            command.ParameterChanged += new CommandParameterChangedEventHandler(Command_ParameterChanged);

            // マウス移動イベントの追加
            command.MouseMove += new CommandMouseMoveEventHandler(Command_MouseMove);

            // 図形の既定値設定変更イベントハンドラ追加
            ShapeSettingsChanged += new ShapeSettingsEventHandler(App_ShapeSettingsChanged);

            // カレント部分図変更イベントハンドラ追加
            CurrentDrawingChanged += new CurrentDrawingChangedEventHandler(App_CurrentDrawingChanged);
        }

        /// <summary>
        /// コマンドの終了時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_End(Object sender, EventArgs e)
        {
            Command command = (Command)sender;

            // 各パラメータのカレントID取得
            GetCommandPrametersCurrentID(command);

            // パラメータ変更イベントの削除
            command.ParameterChanged -= new CommandParameterChangedEventHandler(Command_ParameterChanged);

            // マウス移動イベントの削除
            command.MouseMove -= new CommandMouseMoveEventHandler(Command_MouseMove);

            // 図形の既定値設定変更イベントの削除
            ShapeSettingsChanged -= new ShapeSettingsEventHandler(App_ShapeSettingsChanged);

            // カレント部分図変更イベントの削除
            CurrentDrawingChanged -= new CurrentDrawingChangedEventHandler(App_CurrentDrawingChanged);

            //ラバーバンドクリア
            ClearRubberBand();
        }

        /// <summary>
        /// パラメータの選択値を取得して設定クラスに保持
        /// </summary>
        /// <param name="command"></param>
        private void GetCommandPrametersCurrentID(Command command)
        {
            Debug.Assert(addInSettings != null);

            // 「計算モード」パラメータ
            ListParameterItem calcModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(calcModeParameterId);
            addInSettings.calcModeListParameterItemIndex = calcModeListParameterItem.SelectedIndex;

            // 「係数」パラメータ
            ListParameterItem coefficientListParameterItem = (ListParameterItem)command.GetParameterItemByID(coefficientTypeParameterId);
            addInSettings.coefficientListParameterItemIndex = coefficientListParameterItem.SelectedIndex;

            // 「精度」パラメータ
            ListParameterItem precisionListParameterItem = (ListParameterItem)command.GetParameterItemByID(precisionParameterId);
            addInSettings.precisionListParameterItemIndex = precisionListParameterItem.SelectedIndex;

            // 「丸め」パラメータ
            ListParameterItem roundModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(roundModeParameterId);
            addInSettings.roundModeListParameterItemIndex = roundModeListParameterItem.SelectedIndex;

            // 「小数点以下のゼロ表記」パラメータ
            ListParameterItem zeroVisibleListParameterItem = (ListParameterItem)command.GetParameterItemByID(zeroVisibleParameterId);
            addInSettings.zeroVisibleListParameterItemIndex = zeroVisibleListParameterItem.SelectedIndex;

            // 「3桁区切り文字タイプ」パラメータ
            ListParameterItem separateTextTypeListParameterItem = (ListParameterItem)command.GetParameterItemByID(separateTextTypeParameterId);
            addInSettings.separateTextTypeListParameterItemIndex = separateTextTypeListParameterItem.SelectedIndex;
        }

        /// <summary>
        /// 図形の既定値設定の変更イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_ShapeSettingsChanged(object sender, ShapeSettingsEventArgs e)
        {
            //ラバーバンドクリア
            ClearRubberBand();
        }

        /// <summary>
        /// カレント部分図の変更イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_CurrentDrawingChanged(object sender, CurrentDrawingChangedEventArgs e)
        {
            //ラバーバンドクリア
            ClearRubberBand();
        }

        /// <summary>
        /// ラバーバンドクリア
        /// </summary>
        private void ClearRubberBand()
        {
            if(rubberBandShapes != null)
            {
                rubberBandShapes.Clear();
                rubberBandShapes = null;
            }

            resultShapesVerticalPut = 0;// 結果の文字図形の並べる向き 0:未定,1:縦,2:横
        }
        
        /// <summary>
        /// コマンド実行中にマウスが移動イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Command_MouseMove(Object sender, CommandMouseMoveEventArgs e)
        {
            Command command = (Command)sender;
            // Command.GetTemporaryParameterItemByID を使用してコマンドの一時的なパラメータアイテムを取得します。
            // 一時的なパラメータアイテムからは、すでにパラメータの入力が確定している場合は確定している入力値が、
            // 入力が確定していない場合はマウスの現在位置などの一時的な値が取得できます。

            // 図形Aの取得
            ShapeGroupParameterItem shapeGroupParameterItemA = (ShapeGroupParameterItem)command.GetParameterItemByID(selectShapeAParameterId);
            if(shapeGroupParameterItemA == null)
                return;

            if(shapeGroupParameterItemA.IsEmpty)
                return;

            SelectedShapeCollection selectedShapeCollectionA = shapeGroupParameterItemA.SelectedShapes;

            // 図形Bの取得
            SelectedShapeCollection selectedShapeCollectionB = null;
            ShapeGroupParameterItem shapeGroupParameterItemB = (ShapeGroupParameterItem)command.GetParameterItemByID(selectShapeBParameterId);
            if(shapeGroupParameterItemB != null)
            {

                if(shapeGroupParameterItemB.IsEmpty)
                    return;

                selectedShapeCollectionB = shapeGroupParameterItemB.SelectedShapes;
            }

            // 「計算モード」のパラメータアイテム取得
            ListParameterItem calcModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(calcModeParameterId);
            CalcMode calcMode = (CalcMode)calcModeListParameterItem.SelectedIndex;

            // ラバーバンド用の図形リスト
            // ラバーバンド用図形がない場合
            Point2d tempOrgPoint = new Point2d(0.0, 0.0);// この座標で一旦作る(常にここで作成したラバーバンド用の図形をマウス位置に移動させて表示させる)
            if(rubberBandShapes == null)
            {
                //ラバーバンド表示を削除する
                ClearRubberBand();

                // 「係数」のパラメータアイテム取得
                ListParameterItem coefficientListParameterItem = (ListParameterItem)command.GetParameterItemByID(coefficientTypeParameterId);
                CoefficientType coefficientType = (CoefficientType)coefficientListParameterItem.SelectedIndex;

                // 「精度」のパラメータアイテム取得
                ListParameterItem precisionListParameterItem = (ListParameterItem)command.GetParameterItemByID(precisionParameterId);
                PrecisionMode precisionMode = (PrecisionMode)precisionListParameterItem.SelectedIndex;

                // 「丸め」のパラメータアイテム取得
                ListParameterItem roundModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(roundModeParameterId);
                RoundMode roundMode = (RoundMode)roundModeListParameterItem.SelectedIndex;

                // 「小数点以下のゼロ表記」のパラメータアイテム取得
                ListParameterItem zeroVisibleListParameterItem = (ListParameterItem)command.GetParameterItemByID(zeroVisibleParameterId);
                ZeroVisible zeroVisible = (ZeroVisible)zeroVisibleListParameterItem.SelectedIndex;

                // 「3桁区切り文字タイプ」のパラメータアイテム取得
                ListParameterItem separateTextTypeListParameterItem = (ListParameterItem)command.GetParameterItemByID(separateTextTypeParameterId);
                SeparateTextType separateTextType = (SeparateTextType)separateTextTypeListParameterItem.SelectedIndex;

                 // 図形作成
                List<Shape> shapes = null;
                int verticalPut = 0;// 並び方向 0:未確定、1:縦、2:横
                int stat = CreateCommandShapes(calcMode, selectedShapeCollectionA, selectedShapeCollectionB, coefficientType, precisionMode, roundMode, zeroVisible, separateTextType, tempOrgPoint, out shapes, ref verticalPut);
                if(stat < 0)
                {
                    return;
                }

                rubberBandShapes = new List<Shape>();
                rubberBandShapes.AddRange(shapes);

                // 並び方向保持
                resultShapesVerticalPut = verticalPut;
            }

            // 表の配置点
            PointParameterItem pointParameterItem = (PointParameterItem)command.GetTemporaryParameterItemByID(locatePointParameterId);
            if(pointParameterItem == null)
                return;

            if(pointParameterItem.IsEmpty)
                return;

            Point2d locatePoint = pointParameterItem.Point;

            // 表示させるラバーバンド図形
            List<Shape> thisRubberBandShapes = new List<Shape>();
            foreach(Shape shape in rubberBandShapes)
            {
                // マウスの位置に移動
                Shape shape2 = shape.Clone();

                // 結果の数字が縦方向に並ぶ場合、shapesの縦方向の位置は変わらないのでX方向だけtempOrgPointから移動させる
                Point2d locatePoint2 = tempOrgPoint;
                Debug.Assert(resultShapesVerticalPut != 0);
                if(resultShapesVerticalPut == 1 && calcMode != CalcMode.CalcMode_Total)
                    locatePoint2.X = locatePoint.X;
                else
                    locatePoint2.Y = locatePoint.Y;

                shape2.Transform(tempOrgPoint, 1.0, 1.0, 0.0, locatePoint2);
                thisRubberBandShapes.Add(shape2);
            }

            try
            {
                // ラバーバンド描画
                command.DrawShapeRubberBand(thisRubberBandShapes.ToArray());
            }
            catch
            {
                throw;
            }

        }


        /// <summary>
        /// コマンドパラメータ変更時の処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Command_ParameterChanged(Object sender, CommandParameterChangedEventArgs e)
        {
            //ラバーバンド表示を削除する
            ClearRubberBand();

            Command command = (Command)sender;

             // 「計算モード」のパラメータアイテム取得
             ListParameterItem calcModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(calcModeParameterId);

            int changedParameterId = e.ParameterID;// 変更のあったパラメータID
            // 「計算モード」切り替え時
            if(changedParameterId == calcModeParameterId)
            {
                int selectedIndex = calcModeListParameterItem.SelectedIndex;
                // 「合計」が選ばれたとき
                if(selectedIndex == (int)CalcMode.CalcMode_Total)
                {
                    //「図形B」パラメータ削除
                    ParameterItem selectShapeBParameterItem = command.GetParameterItemByID(selectShapeBParameterId);
                    if(selectShapeBParameterItem != null)
                    {
                        CommandParameterGroup selectShapeBParameterGroup = selectShapeBParameterItem.Group;
                        command.ParameterGroups.Remove(selectShapeBParameterGroup);
                    }
                }
                else
                {
                    //「図形A」パラメータの後に「図形2」パラメータ追加
                    ParameterItem selectShapeAParameterItem = command.GetParameterItemByID(selectShapeAParameterId);
                    int selectShapeAParameterItemIndex = selectShapeAParameterItem.Index;

                    command.ParameterGroups.Insert(selectShapeAParameterItemIndex + 1, ParameterItemType.ShapeGroup, selectShapeBParameterId, Properties.Resources.ShapeBGroupParameterName);
                    ShapeGroupParameterItem shape2ParameterItem = (ShapeGroupParameterItem)command.GetParameterItemByID(selectShapeBParameterId);
                    shape2ParameterItem.ShapeTypeFilter = ShapeTypeFilters.Text;
                    shape2ParameterItem.Description = Properties.Resources.ShapeBGroupParameterDescription;
                }
            }

            //ドキュメントの取得
            Document doc = ActiveDocument;
            Drawing drawing = doc.CurrentDrawing;
 

            // 図形Aの取得
            ShapeGroupParameterItem shapeGroupParameterItemA = (ShapeGroupParameterItem)command.GetParameterItemByID(selectShapeAParameterId);
            if(shapeGroupParameterItemA == null)
                return;

            if(shapeGroupParameterItemA.IsEmpty)
                return;

            SelectedShapeCollection selectedShapeCollectionA = shapeGroupParameterItemA.SelectedShapes;

            // 図形Bの取得
            SelectedShapeCollection selectedShapeCollectionB = null;
            ShapeGroupParameterItem shapeGroupParameterItemB = (ShapeGroupParameterItem)command.GetParameterItemByID(selectShapeBParameterId);
            if(shapeGroupParameterItemB != null)
            {
                if(shapeGroupParameterItemB.IsEmpty)
                    return;

                selectedShapeCollectionB = shapeGroupParameterItemB.SelectedShapes;
            }

            // 「計算モード」
            CalcMode calcMode = (CalcMode)calcModeListParameterItem.SelectedIndex;

            // 「係数」のパラメータアイテム取得
            ListParameterItem coefficientListParameterItem = (ListParameterItem)command.GetParameterItemByID(coefficientTypeParameterId);
            CoefficientType coefficientType = (CoefficientType)coefficientListParameterItem.SelectedIndex;

            // 「精度」のパラメータアイテム取得
            ListParameterItem precisionListParameterItem = (ListParameterItem)command.GetParameterItemByID(precisionParameterId);
            PrecisionMode precisionMode = (PrecisionMode)precisionListParameterItem.SelectedIndex;

            // 「丸め」のパラメータアイテム取得
            ListParameterItem roundModeListParameterItem = (ListParameterItem)command.GetParameterItemByID(roundModeParameterId);
            RoundMode roundMode = (RoundMode)roundModeListParameterItem.SelectedIndex;

            // 「小数点以下のゼロ表記」のパラメータアイテム取得
            ListParameterItem zeroVisibleListParameterItem = (ListParameterItem)command.GetParameterItemByID(zeroVisibleParameterId);
            ZeroVisible zeroVisible = (ZeroVisible)zeroVisibleListParameterItem.SelectedIndex;

            // 「3桁区切り文字タイプ」のパラメータアイテム取得
            ListParameterItem separateTextTypeListParameterItem = (ListParameterItem)command.GetParameterItemByID(separateTextTypeParameterId);
            SeparateTextType separateTextType = (SeparateTextType)separateTextTypeListParameterItem.SelectedIndex;

            // 結果の配置点
            PointParameterItem pointParameterItem = (PointParameterItem)command.GetParameterItemByID(locatePointParameterId);
            if(pointParameterItem == null)
                return;

            if(pointParameterItem.IsEmpty)
                return;

            Point2d locatePoint = pointParameterItem.Point;

            // 図形作成
            List<Shape> shapes = null;
            int verticalPut = 0; // 並び方向 0:未確定、1:縦、2:横
            int stat = CreateCommandShapes(calcMode, selectedShapeCollectionA, selectedShapeCollectionB, coefficientType, precisionMode, roundMode, zeroVisible, separateTextType, locatePoint, out shapes, ref verticalPut);
            if(stat < 0)
            {
                ShowErrMessage(stat);

                e.CommandState = CommandState.ParameterError;
                return;
            }

            // UNDO 可能な操作を開始
            doc.UndoManager.BeginUndoUnit();

            // 図面上にプロット
            foreach(Shape shape in shapes)
            {
                // 描画図形追加
                drawing.Shapes.Add(shape);
            }

            // UNDO 可能な操作を終了
            doc.UndoManager.EndUndoUnit();

            e.CommandState = CommandState.Execute;
        }

        /// <summary>
        /// 出力図形の作成
        /// </summary>
        /// <param name="calcMode">計算モード</param>
        /// <param name="selectedShapeCollectionA">選択図形A</param>
        /// <param name="selectedShapeCollectionB">選択図形B</param>
        /// <param name="coefficientType">係数タイプ</param>
        /// <param name="precisionMode">精度モード</param>
        /// <param name="roundMode">丸めモード</param>
        /// <param name="zeroVisible">小数点以下のゼロ表記モード</param>
        /// <param name="separateTextType">3桁区切り文字タイプ</param>
        /// <param name="locatePoint">配置点</param>
        /// <param name="shapes">出力図形</param>
        /// <param name="verticalPut">並び方向 0:未確定、1:縦、2:横</param>
        /// <returns></returns>
        private int CreateCommandShapes(CalcMode calcMode, SelectedShapeCollection selectedShapeCollectionA, SelectedShapeCollection selectedShapeCollectionB,
                                        CoefficientType coefficientType, PrecisionMode precisionMode, RoundMode roundMode, ZeroVisible zeroVisible, SeparateTextType separateTextType, Point2d locatePoint, 
                                        out List<Shape> shapes, ref int verticalPut)
        {
            shapes = new List<Shape>();// 作成図形

            //ドキュメントの取得
            Document doc = ActiveDocument;
            Drawing drawing = doc.CurrentDrawing;

            // 選択図形の数
            int selectedShapeCollectionACount = 0;
            int selectedShapeCollectionBCount = 0;
            selectedShapeCollectionACount = selectedShapeCollectionA.Count;
            // 「合計」以外の場合
            if(calcMode != CalcMode.CalcMode_Total)
            {
                Debug.Assert(selectedShapeCollectionB != null);
                selectedShapeCollectionBCount = selectedShapeCollectionB.Count;
                if(selectedShapeCollectionACount != selectedShapeCollectionBCount)
                    return RC_ERR_DIFFERENT_SELECTEDSHAPES_COUNT;// エラー：選択した図形Aと図形Bの数が異なります。
            }

            if(selectedShapeCollectionACount == 0)
                return RC_ERR_NON_SELECTEDSHAPES;// エラー：選択図形がありません。

            // 部分図チェック
            // カレント部分図のものだけで計算する
            int drawingId = drawing.ID;
            if(!IsAllSameDrawingIDShapes(selectedShapeCollectionA, ref drawingId))
                return RC_ERR_NOT_CURRENTDRAWINGSHAPE;// エラー：カレント部分図以外の図形が選ばれています。

            if(selectedShapeCollectionBCount > 0)
                if(!IsAllSameDrawingIDShapes(selectedShapeCollectionB, ref drawingId))
                    return RC_ERR_NOT_CURRENTDRAWINGSHAPE;// エラー：カレント部分図以外の図形が選ばれています。

            // 選択した文字図形をデータ順ではなく、見た目に縦か横に並べた順に並べる
            // 並びが横方向か縦方向かも決める
            // A
            List<TextShape> textShapeListA = new List<TextShape>();
            int stat = GetSortedTextShapeList(selectedShapeCollectionA, textShapeListA, ref verticalPut);
            if(stat < 0)
                return stat;

            // B
            List<TextShape> textShapeListB = new List<TextShape>();
            if(selectedShapeCollectionBCount > 0)
            {
                // 向きはAと同じで
                stat = GetSortedTextShapeList(selectedShapeCollectionB, textShapeListB, ref verticalPut);
                if(stat < 0)
                    return stat;
            }

            // 結果の数値リスト作成
            List<double> resultValues = new List<double>();
            int maxDecimalsDigits = 0;// 選択数値文字の最大小数部分桁数
            // 「合計」の場合
            if(calcMode == CalcMode.CalcMode_Total)
            {
                // 選択図形Aから数値の合計値取得
                double totalValue = 0.0;
                stat = GetSumValue(textShapeListA, out totalValue, out maxDecimalsDigits);
                if(stat < 0)
                    return stat;

                resultValues.Add(totalValue);
            }
            else
            {
                // 他の計算モードの計算結果
                stat = GetCalcResultValues(calcMode, textShapeListA, textShapeListB, resultValues, out maxDecimalsDigits);
                if(stat < 0)
                    return stat;
            }

            // 書込みレイヤ
            Layer currentLayer = doc.LayerTable.CurrentLayer;

            // 有効数字
            int valueDecimals = (int)precisionMode;
            // 選択数値図形で最も小数部分が多いものに合わせる場合
            if(precisionMode == PrecisionMode.PrecisionMode_F)
                valueDecimals = maxDecimalsDigits;

            // 文字作成
            TextShape firstSelectTextShapeA = textShapeListA[0];
            int resultValuesCount = resultValues.Count;
            Debug.Assert(resultValuesCount <= textShapeListA.Count);
            for(int i = 0; i < resultValuesCount; i++)
            {
                double value = resultValues[i];
                if(coefficientType == CoefficientType.CoefficientType_0_001)
                    value *= 0.001;
                else if(coefficientType == CoefficientType.CoefficientType_100)
                    value *= 100.0;

                double resultValue = 0.0;

                // 丸め
                // 四捨五入
                if(roundMode == RoundMode.Round_45)
                    resultValue = Math.Round(value, valueDecimals, MidpointRounding.AwayFromZero);
                //「切り上げ」
                else if(roundMode == RoundMode.Round_Up)
                    resultValue = RoundUp(value, valueDecimals);
                //「切り捨て」
                else if(roundMode == RoundMode.Round_Cut)
                    resultValue = Truncate(value, valueDecimals);

                // 数値から文字変換
                stat = ConvertValueToText(resultValue, valueDecimals, precisionMode, roundMode, zeroVisible, separateTextType, out string resultText);
                if(stat < 0)
                    return stat;

                // 文字図形作成
                // 選択図形Aの並びに合わせる
                TextShape selectTextShapeA = textShapeListA[i];
                Point2d[] boundingBoxPoints = selectTextShapeA.GetBoundingBox();// 範囲矩形の右下左上の2点を取得
                double locatePointX = locatePoint.X;
                double locatePointY = locatePoint.Y;
                // 縦に並べる場合
                if(verticalPut == 1 && calcMode != CalcMode.CalcMode_Total)
                {
                    locatePointY = boundingBoxPoints[1].Y;
                }
                // 「合計」モードか横に並べる場合
                else //if(verticalPut == 2)
                {
                    locatePointX = boundingBoxPoints[1].X;
                }

                // 配置点
                Point2d textLocatePoint = new Point2d(locatePointX, locatePointY);

                // 図形の作成
                TextShape textShape = ShapeFactory.CreateText(resultText, textLocatePoint, 0.0);

                // 設定
                doc.TextSettings.CopyToShape(textShape);// 先ず既定値をセット
                // フォントの高さや種類、文字間隔などは選択図形の最初のものと同じにする
                textShape.FontName = firstSelectTextShapeA.FontName;
                textShape.FontHeight = firstSelectTextShapeA.FontHeight;
                textShape.ColorNumber = firstSelectTextShapeA.ColorNumber;
                textShape.Spacing = firstSelectTextShapeA.Spacing;
                textShape.WidthScale = firstSelectTextShapeA.WidthScale;
                textShape.Slant = firstSelectTextShapeA.Slant;
                textShape.Bold = firstSelectTextShapeA.Bold;

                // 配置基準は右上
                textShape.Alignment = Alignment.TopRight;

                // レイヤ
                ((Shape)textShape).Layer = currentLayer;

                // 図形追加
                shapes.Add(textShape);
            }

            return RC_ERR_SUCCESS;
        }

        /// <summary>
        /// 値の小数部を切り捨て、指定の桁数に丸める
        /// </summary>
        /// <param name="value">値</param>
        /// <param name="decimals">小数部桁数</param>
        /// <returns>丸めた値</returns>
        static public double Truncate(double value, int decimals)
        {
            // 小数部桁数の10の累乗を取得
            double pow = Math.Pow(10, decimals);
            return 0 <= value ? Math.Floor(value * pow) / pow : Math.Ceiling(value * pow) / pow;
        }

        /// <summary>
        /// 値の小数部を切り上げ、指定の桁数に丸める
        /// </summary>
        /// <param name="value">値</param>
        /// <param name="decimals">小数部桁数</param>
        /// <returns>丸めた値</returns>
        static public double RoundUp(double value, int decimals)
        {
            // 小数部桁数の10の累乗を取得
            double pow = Math.Pow(10, decimals);
            return 0 <= value ? Math.Ceiling(value * pow) / pow : Math.Floor(value * pow) / pow;
        }


        /// <summary>
        /// 全て同じ部分図の図形か
        /// </summary>
        /// <param name="shapeCollection"></param>
        /// <param name="drawingId">　-1:指定なし。参照渡しで最初の図形の部分図IDが返る</param>
        /// <returns></returns>
        private bool IsAllSameDrawingIDShapes(SelectedShapeCollection shapeCollection, ref int drawingId)
        {
            foreach(SelectedShape selectedShape in shapeCollection)
            {
                Shape shape = selectedShape.Shape;

                if(drawingId < 0)
                {
                    drawingId = shape.Drawing.ID;
                    continue;
                }

                if(drawingId != shape.Drawing.ID)
                    return false;
            }

            return true;
        }

        /// <summary>
        /// 選択した文字図形を見た目に縦か横に並べた順に並べる
        /// </summary>
        /// <param name="selectedShapeCollection"></param>
        /// <param name="textShapeList"></param>
        /// <param name="verticalPut">並び方向 0:未確定、1:縦(要素が一つのときも)、2:横</param>
        /// <returns></returns>
        private int GetSortedTextShapeList(SelectedShapeCollection selectedShapeCollection, List<TextShape> textShapeList, ref int verticalPut)
        {
            textShapeList.Clear();
            
            // コレクションから配置角度と書かれる向きのチェックをしながらリストに変換
            foreach(SelectedShape selectedShape in selectedShapeCollection)
            {
                TextShape textShape = (TextShape)selectedShape.Shape;

                // 配置角度チェック
                if(!(textShape.Angle <= RcPrecisionRadianAngular))
                    return RC_ERR_TEXT_LOCATE_ANGLE;

                // 縦書きチェック
                if(textShape.DirectionVertical)
                    return RC_ERR_TEXT_DIRECTIONVERTICAL;

                if(textShape is TextShape)
                {
                    textShapeList.Add(textShape);
                }
                // 文字以外 エラー
                else
                {
                    return RC_ERR_SELECTED_NONTEXT;// エラー：文字以外が選択されていた。
                }
            }

            // 並べる向きを決める
            // 1つ目とそれ以降の位置関係で縦並びか横並びか決める
            int textShapeListCount = textShapeList.Count;
            Debug.Assert(textShapeListCount > 0);

            TextShape textShape1 = textShapeList[0];// 1つ目

            // 矩形範囲のXY最大最小中央値取得
            GetMaxMinMiddleXY(textShape1, out double minX1, out double maxX1, out double middleX1, out double minY1, out double maxY1, out double middleY1);

            // 並び方向の決定&チェック
            // 未定義なら初期値セット
            int orgVerticalPut = verticalPut;
            if(verticalPut == 0)
                verticalPut = 1;//　初期値：縦(1),  横(2), 定まらない(-1)

            int tempVerticalPut = 1;
            for(int i = 1; i < textShapeListCount; i++)
            {
                TextShape textShape2 = textShapeList[i];
                GetMaxMinMiddleXY(textShape2, out double minX2, out double maxX2, out double middleX2, out double minY2, out double maxY2, out double middleY2);

                // 1番目の矩形の中心X座標が2番目以降の矩形のX最小最大間にあれば縦に並んでいる
                // あるいは逆に2番目以降の矩形の中心X座標が1番目の矩形のX最小最大間にあれば縦に並んでいる
                if((minX2 <= middleX1 && middleX1 <= maxX2) || (minX1 <= middleX2 && middleX2 <= maxX1))
                    tempVerticalPut = 1;
                // 1番目の矩形の中心Y座標が2番目以降の矩形のY最小最大間にあれば横に並んでいる
                // あるいは逆に2番目以降の矩形の中心Y座標が1番目の矩形のY最小最大間にあれば縦に並んでいる
                else if((minY2 <= middleY1 && middleY1 <= maxY2) || (minY1 <= middleY2 && middleY2 <= maxY1))
                    tempVerticalPut = 2;
                // それ以外は並びが整列していないと判断する
                else
                    tempVerticalPut = -1;

                // エラー：選択した文字図形は整列していません。
                if(tempVerticalPut < 0)
                    return RC_ERR_TEXT_NOT_ALIGNED;

                // 指定された並びと異なる
                if(orgVerticalPut > 0 && tempVerticalPut != orgVerticalPut)
                    return RC_ERR_TEXT_NOT_ALIGNED;

                // 途中で向きが変わった
                if(i > 1 && tempVerticalPut != verticalPut)
                    return RC_ERR_TEXT_NOT_ALIGNED;

                verticalPut = tempVerticalPut;
            }

            // 配置点のY座標値でソート(縦方向に並べる Yの+方向から-方向に)
            if(verticalPut == 1)
                textShapeList.Sort((a, b) => Math.Sign(b.Point.Y - a.Point.Y));
            // 配置点のX座標値でソート(横方向に並べる Xの-方向から+方向に)
            else if(verticalPut == 2)
                textShapeList.Sort((a, b) => Math.Sign(a.Point.Y - b.Point.Y));
            else
            {
                Debug.Assert(verticalPut == 1 || verticalPut == 2);
            }

            return RC_ERR_SUCCESS;
        }

        /// <summary>
        /// 図形範囲矩形のXY最大最小中央値取得
        /// </summary>
        /// <param name="shape"></param>
        /// <param name="minX"></param>
        /// <param name="maxX"></param>
        /// <param name="middleX"></param>
        /// <param name="minY"></param>
        /// <param name="maxY"></param>
        /// <param name="middleY"></param>
        public void GetMaxMinMiddleXY(Shape shape, out double minX, out double maxX, out double middleX, out double minY, out double maxY, out double middleY)
        {
            // 範囲矩形の右下左上の2点を取得
            Point2d[] boundingBoxPoints = shape.GetBoundingBox();

            minX = boundingBoxPoints[0].X;
            maxX = boundingBoxPoints[1].X;
            middleX = (maxX + minX) * 0.5;
            minY = boundingBoxPoints[0].Y;
            maxY = boundingBoxPoints[1].Y;
            middleY = (maxY + minY) * 0.5;
        }

        /// <summary>
        /// 選択図形から数値の合計値取得
        /// </summary>
        /// <param name="valueTextShapes">選択数値文字図形</param>
        /// <param name="totalValue">合計値</param>
        /// <param name="maxDecimalsDigits">選択数値文字の最大小数部分桁数</param>
        /// <returns></returns>
        private int GetSumValue(List<TextShape> valueTextShapes, out double totalValue, out int maxDecimalsDigits)
        {
            // 合計値
            totalValue = 0.0;

            // 選択数値文字の最大小数部分桁数
            maxDecimalsDigits = 0;
            
            foreach(Shape shape in valueTextShapes)
            {
                // 文字から数値にして足し算
                if(shape is TextShape)
                {
                    string valueText = ((TextShape)shape).Text;
                    valueText = valueText.TrimEnd();

                    // 文字列をdouble型に変換
                    if(!double.TryParse(valueText, out double value))
                        return RC_ERR_CONVERT_TEXT_TONUM;// エラー：数値変換できなかった

                    totalValue += value;

                    // 数値文字の最大小数部分桁数更新
                    UpdateMaxDecimalsDigits(valueText, ref maxDecimalsDigits);
                }
                // 文字以外 エラー
                else
                {
                    return RC_ERR_SELECTED_NONTEXT;// エラー：文字以外が選択されていた。
                }
            }

            return RC_ERR_SUCCESS;
        }

        /// <summary>
        /// 数値文字の最大小数部分桁数更新
        /// </summary>
        /// <param name="valueText"></param>
        /// <param name="maxDecimalsDigits"></param>
        private void UpdateMaxDecimalsDigits(string valueText, ref int maxDecimalsDigits)
        {
            // 数値文字の最大小数部分桁数
            // 整数部分と小数部分に分ける
            string[] part = valueText.Split('.');
            if(part.Length > 1)
            {
                string decimalsText = part[1];
                int decimalsTextLength = decimalsText.Length;
                if(decimalsTextLength > maxDecimalsDigits)
                    maxDecimalsDigits = decimalsTextLength;
            }
        }

        /// <summary>
        /// 2つのリストの同じ順番の数字を指定したモードで計算して結果のリストを取得
        /// </summary>
        /// <param name="calcMode"></param>
        /// <param name="textShapeListA"></param>
        /// <param name="textShapeListB"></param>
        /// <param name="resultValues">計算結果数値リスト</param>
        /// <param name="maxDecimals">選択数値文字の最大小数部分桁数</param>
        /// <returns></returns>
        private int GetCalcResultValues(CalcMode calcMode, List<TextShape> textShapeListA, List<TextShape> textShapeListB, List<double> resultValues, out int maxDecimalsDigits)
        {
            resultValues.Clear();

            // 選択数値文字の最大小数部分桁数
            maxDecimalsDigits = 0;

            Debug.Assert(calcMode == CalcMode.CalcMode_Addition || calcMode == CalcMode.CalcMode_Subtraction || calcMode == CalcMode.CalcMode_Multiplication || calcMode == CalcMode.CalcMode_Division);

            // 図形Aと図形Bの選択数は同じであること
            int textShapesACount = textShapeListA.Count;
            int textShapesBCount = textShapeListB.Count;
            Debug.Assert(textShapesACount > 0 && textShapesACount == textShapesBCount);
            if(textShapesACount != textShapesBCount)
                return RC_ERR_DIFFERENT_SELECTEDSHAPES_COUNT;// エラー：選択した図形Aと図形Bの数が異なります。

            for(int i = 0; i < textShapesACount; i++)
            {
                // A
                string valueTextA = textShapeListA[i].Text;
                valueTextA = valueTextA.TrimEnd();

                // 文字Aをdouble型に変換
                if(!double.TryParse(valueTextA, out double valueA))
                    return RC_ERR_CONVERT_TEXT_TONUM;// エラー：数値変換できなかった

                // 数値文字の最大小数部分桁数更新
                UpdateMaxDecimalsDigits(valueTextA, ref maxDecimalsDigits);

                // B
                string valueTextB = textShapeListB[i].Text;
                valueTextB = valueTextB.TrimEnd();

                // 文字Bをdouble型に変換
                if(!double.TryParse(valueTextB, out double valueB))
                    return RC_ERR_CONVERT_TEXT_TONUM;// エラー：数値変換できなかった

                // 数値文字の最大小数部分桁数更新
                UpdateMaxDecimalsDigits(valueTextB, ref maxDecimalsDigits);

                // 計算
                double resultValue = 0.0;
                switch(calcMode)
                {
                    // 足し算
                    case CalcMode.CalcMode_Addition:
                        resultValue = valueA + valueB;
                        break;
                    // 引き算
                    case CalcMode.CalcMode_Subtraction:
                        resultValue = valueA - valueB;
                        break;
                    // 掛け算
                    case CalcMode.CalcMode_Multiplication:
                        resultValue = valueA * valueB;
                        break;
                    // 割り算
                    case CalcMode.CalcMode_Division:
                        
                        // エラー:ゼロ割
                        if(Math.Abs(valueB) <= RcPrecisionConfusion)
                            return RC_ERR_ZERO_DIVISION;

                        resultValue = valueA / valueB;
                        break;
                }

                resultValues.Add(resultValue);
            }

            return RC_ERR_SUCCESS;
        }

        /// <summary>
        /// 2点が同じ座標か
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <returns></returns>
        public Boolean IsPointEquals(Point2d point1, Point2d point2)
        {
            return Geometry.IsPointEquals(point1, point2);
        }

        /// <summary>
        /// 2点間の中心座標取得
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <returns></returns>
        public Point2d CenterPoint(Point2d point1, Point2d point2)
        {
            Point2d centerPoint = Geometry.GetMiddlePoint(point1, point2);
            return centerPoint;
        }

        /// <summary>
        /// 水平に対する2点のなす角度取得
        /// </summary>
        /// <param name="point1">基点</param>
        /// <param name="point2">方向点</param>
        /// <returns> 2点のなす角度(Degree) </returns>
        public double Get2PointsAngle(Point2d point1, Point2d point2)
        {
             return Geometry.GetAngle(point1, point2);
        }

        /// <summary>
        /// 数値を指定した書式の文字に変換する
        /// </summary>
        /// <param name="value"></param>
        /// <param name="valueDecimals"></param>
        /// <param name="precisionMode"></param>
        /// <param name="roundMode"></param>
        /// <param name="zeroVisible"></param>
        /// <param name="separateTextType"></param>
        /// <param name="resultText"></param>
        /// <returns></returns>
        private int ConvertValueToText(double value, int valueDecimals, PrecisionMode precisionMode, RoundMode roundMode, ZeroVisible zeroVisible, SeparateTextType separateTextType, out string resultText)
        { 
            // 書式 指定した精度で0詰め
            string valueStringFormatParamf = "{0:f" + valueDecimals.ToString() + "}";

            // 数値から文字変換
            string text = String.Format(valueStringFormatParamf, value);
                
            // 整数部分と小数部分に分ける
            string[] part = text.Split('.');

            // 整数部分
            string partText1 = part[0];

            // 3桁区切り文字(一旦数値にしてからFormat書式指定で行ってもよいがスペース区切りもあるので文字のまま)
            int partText1Length = partText1.Length;
            if(partText1Length > 3 && separateTextType != SeparateTextType.SeparateTextType_Non)
            {
                string commaString = ",";
                if(separateTextType == SeparateTextType.SeparateTextType_Space)
                    commaString = " ";

                string withCommaString = "";
                int counter = 0;
                for(int j = partText1Length - 1; j >= 0; j--)
                {
                    if(counter == 3)
                    {
                        withCommaString = withCommaString.Insert(0, commaString);
                        counter = 0;
                    }

                    withCommaString = withCommaString.Insert(0, partText1[j].ToString());

                    counter++;
                }

                partText1 = withCommaString;// 入れ替え
            } 

            // 小数部分
            string partText2 = "";
            if(part.Length > 1)
            {
                partText2 = part[1];

                // 小数点以下のゼロ表記がオフの場合は、0の代わりにスペースを追加
                // 右寄せで小数点位置が揃うように
                if(zeroVisible == ZeroVisible.ZeroVisible_Hide)
                {
                    int oldTextLength = partText2.Length;
                    partText2 = partText2.TrimEnd('0');// 右の0を削除
                    int textLength = partText2.Length;
                    int delTextCount = oldTextLength - textLength;
                    for(int j = 0; j < delTextCount; j++)
                        partText2 += " ";
                }
            }

            // 出力文字列
            // 整数部分
            string text2 = partText1;
            // 小数部分の追加
            string trimedPpartText2 = partText2.Trim();
            if(trimedPpartText2.Length > 0)
            {
                text2 += "." + partText2;
            }
            // 精度が0でなく小数部分がない場合
            else if(valueDecimals > 0)
            {
                for(int j = 0; j < valueDecimals + 1; j++)
                    text2 += " ";// 小数点の位置と精度に合わせた小数部分にスペースを追加する
            }

            resultText = text2;

            return RC_ERR_SUCCESS;
        }


        /// <summary>
        /// エラーメッセージの表示
        /// </summary>
        /// <param name="errorNumber">エラー番号</param>
        private void ShowErrMessage(int errorNumber)
        {
            string strErr = "";
            switch(errorNumber)
            {
                // 選択した図形Aと図形Bの数が異なります。
                case RC_ERR_DIFFERENT_SELECTEDSHAPES_COUNT:
                    strErr = Properties.Resources.ErrDifferentSelectedShapesCount;
                    break;
                // 選択図形がありません。
                case RC_ERR_NON_SELECTEDSHAPES:
                    strErr = Properties.Resources.ErrNonSelectedShapes;
                    break;
                // カレント部分図以外の図形が選ばれています。
                case RC_ERR_NOT_CURRENTDRAWINGSHAPE:
                    strErr = Properties.Resources.ErrNotCurrentDrawingShape;
                    break;
                // 文字以外が選ばれています。
                case RC_ERR_SELECTED_NONTEXT:
                    strErr = Properties.Resources.ErrSelectedNonText;
                    break;
                // 文字を数値に変換できませんでした。
                case RC_ERR_CONVERT_TEXT_TONUM:
                    strErr = Properties.Resources.ErrConvertTextToNum;
                    break;
                // 選択した図形Aと図形Bの並びの向きが異なります。
                case RC_ERR_DIFFERENT_TEXT_PUT:
                    strErr = Properties.Resources.ErrDifferentTextPut;
                    break;
                // 選択した図形Bに0があります。(割り算実行時)
                case RC_ERR_ZERO_DIVISION:
                    strErr = Properties.Resources.ErrZeroDivision;
                    break;
                // 選択した文字図形に配置角度が0以外のものが含まれています。
                case RC_ERR_TEXT_LOCATE_ANGLE:
                    strErr = Properties.Resources.ErrTextLocateAngle;
                    break;
                // 選択した文字図形に縦書きが含まれています。
                case RC_ERR_TEXT_DIRECTIONVERTICAL:
                    strErr = Properties.Resources.ErrTextDirectionVertical;
                    break;
                // 選択した文字図形は整列していません。
                case RC_ERR_TEXT_NOT_ALIGNED:
                    strErr = Properties.Resources.ErrTextNotAligned;
                    break;
            }

            if(strErr != "")
                MessageBox.Show(strErr, AddInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);

        }
    }
}
